#include "seval.h"
#include "Pico.h"
#include "escheme.h"

static Scheme_Env *_scheme_environment_;
static Scheme_Object *exn_catching_apply, *exn_p, *exn_message;


static void init_exn_catching_apply() {
  if (!exn_catching_apply) {
    char *e = 
      "(lambda (thunk) "
	"(with-handlers ([void (lambda (exn) (cons #f exn))]) "
	  "(cons #t (thunk))))";

    scheme_register_extension_global(&exn_catching_apply, sizeof(Scheme_Object *));
    scheme_register_extension_global(&exn_p, sizeof(Scheme_Object *));
    scheme_register_extension_global(&exn_message, sizeof(Scheme_Object *));
    
    exn_catching_apply = scheme_eval_string(e, _scheme_environment_);
    exn_p = scheme_lookup_global(scheme_intern_symbol("exn?"), _scheme_environment_);
    exn_message = scheme_lookup_global(scheme_intern_symbol("exn-message"), _scheme_environment_);
  }
}

/* This function applies a thunk, returning the Scheme value if there's no exception, 
   otherwise returning NULL and setting *exn to the raised value (usually an exn 
   structure). */
static Scheme_Object *_apply_thunk_catch_exceptions(Scheme_Object *f, Scheme_Object **exn)
{
  Scheme_Object *v;

  init_exn_catching_apply();
  
  v = _scheme_apply(exn_catching_apply, 1, &f);
  /* v is a pair: (cons #t value) or (cons #f exn) */

  if (SCHEME_TRUEP(SCHEME_CAR(v)))
    return SCHEME_CDR(v);
  else {
    *exn = SCHEME_CDR(v);
    return NULL;
  }
}

static Scheme_Object *extract_exn_message(Scheme_Object *v) {
  init_exn_catching_apply();

  if (SCHEME_TRUEP(_scheme_apply(exn_p, 1, &v)))
    return _scheme_apply(exn_message, 1, &v);
  else
    return NULL; /* Not an exn structure */
}


static Scheme_Object *do_eval(void *s, int noargc, Scheme_Object **noargv) {
  return scheme_eval_string((char *)s, _scheme_environment_);
}

static Scheme_Object *eval_string_or_get_exn_message(char *s) {
  Scheme_Object *v, *exn;

  v = _apply_thunk_catch_exceptions(scheme_make_closed_prim(do_eval, s), &exn);
  /* Got a value? */
  if (v)
    return v;

  v = extract_exn_message(exn);
  /* Got an exn? */
  if (v) {
    display_in_scheme(SCHEME_STR_VAL(v));
	return v;
  }
  
  /* `raise' was called on some arbitrary value */
  return exn;
}

void _eval_in_scheme_(const char* s) {
  eval_string_or_get_exn_message(s);
}

void _init_scheme_() {
  _scheme_environment_ = (Scheme_Env *)scheme_make_namespace(0, NULL);
  scheme_dont_gc_ptr(_scheme_environment_);
}
